import numpy as np
from qosm_core import Vec3

from qosm.utils.Field import Field
from qosm.waves.VAGB import VAGB
from qosm.utils.Pose import Vector
from qosm.sources.Source import Source


class GaussianBeam(Source):
    TYPE = '_GB_'

    def __init__(self, frequency_GHz: float):
        super(GaussianBeam, self).__init__(frequency_GHz)
        self._w0 = 0
        self._z0 = 0
        self._phase0 = 0
        self._alpha = [0, 0]

    def init(self, **kwargs):
        self._w0 = kwargs['w0']
        self._z0 = kwargs['z0']
        self._alpha = kwargs['polarisation']

    def e_field(self, r_pts_src: Vector, limit_pw: float = None, compute_power: bool = True, pbar=None):
        beam = VAGB(
            polarisation=self._alpha,
            k0=self.k0,
            ior=(1., 0.),
            w0=[self._w0, self._w0],
            z0=[self._z0, self._z0]
        )
        # removed phase offset
        E, _ = beam.compute_fields(r_pts_src)
        E *= beam.Z
        return E

    def h_field(self, r_pts_src: Vector, limit_pw: float = None, compute_power: bool = True, pbar=None):
        beam = VAGB(
            polarisation=self._alpha,
            k0=self.k0,
            ior=(1., 0.),
            w0=[self._w0, self._w0],
            z0=[self._z0, self._z0]
        )
        # removed phase offset
        _, H = beam.compute_fields(r_pts_src)
        return H

    def fields(self, r_pts_src: Vector, pbar=None):
        beam = VAGB(
            polarisation=self._alpha,
            k0=self.k0,
            ior=(1., 0.),
            w0=[self._w0, self._w0],
            z0=[self._z0, self._z0]
        )
        # removed phase offset
        E, H = beam.compute_fields(r_pts_src)

        # Poynting vector
        P = Vector(N=E.shape[0])
        Hc = H.conj()
        P[:, 0] = 0.5 * np.real(E[:, 1] * Hc[:, 2] - E[:, 2] * Hc[:, 1])
        P[:, 1] = 0.5 * np.real(E[:, 2] * Hc[:, 0] - E[:, 0] * Hc[:, 2])
        P[:, 2] = 0.5 * np.real(E[:, 0] * Hc[:, 1] - E[:, 1] * Hc[:, 0])

        return E, H, P

    def oriented_fields(self, r_pts_src: Vector, z_probe_src: Vec3, pbar=None):
        beam = VAGB(
            polarisation=self._alpha,
            k0=self.k0,
            ior=(1., 0.),
            w0=[self._w0, self._w0],
            z0=[self._z0, self._z0]
        )
        # removed phase offset
        E, H = beam.compute_fields(r_pts_src)
        E *= Field.Z0
        H *= Field.Z0

        # Poynting vector
        P = Vector(N=E.shape[0])
        Hc = H.conj()
        P[:, 0] = 0.5 * np.real(E[:, 1] * Hc[:, 2] - E[:, 2] * Hc[:, 1])
        P[:, 1] = 0.5 * np.real(E[:, 2] * Hc[:, 0] - E[:, 0] * Hc[:, 2])
        P[:, 2] = 0.5 * np.real(E[:, 0] * Hc[:, 1] - E[:, 1] * Hc[:, 0])

        if z_probe_src is not None:
            s = P[:, 0]*z_probe_src[0] + P[:, 1]*z_probe_src[1] + P[:, 2]*z_probe_src[2]
            E[s < 0] = 0.
            H[s < 0] = 0.
            P[s < 0] = 0.

        return E, H, P
